import pytest
from django_dynamic_fixture import G

from silver.models import Transaction, Proforma, Invoice, Customer
from silver import payment_processors
from silver_payu.models import PayUPaymentMethod


@pytest.fixture()
def customer():
    return G(Customer, currency='RON',
             address_1=u"Zalaegerszegi utca 65.\nH-5484 ligetv\xc3\xa1r", address_2='9',
             sales_tax_number=0)


@pytest.fixture()
def payment_processor():
    return payment_processors.get_instance('payu_manual')


@pytest.fixture()
def payment_processor_triggered():
    return payment_processors.get_instance('payu_triggered')


@pytest.fixture()
def payment_processor_triggered_v2():
    return payment_processors.get_instance('payu_triggered_v2')


@pytest.fixture()
def payment_method(customer, payment_processor):
    return G(PayUPaymentMethod, customer=customer,
             payment_processor=payment_processor.name)


@pytest.fixture()
def payment_method_triggered(customer, payment_processor_triggered):
    return G(PayUPaymentMethod, customer=customer,
             payment_processor=payment_processor_triggered.name)


@pytest.fixture()
def payment_method_triggered_v2(customer, payment_processor_triggered_v2):
    return G(PayUPaymentMethod, customer=customer,
             payment_processor=payment_processor_triggered_v2.name)


@pytest.fixture()
def proforma(customer):
    return G(Proforma, state=Invoice.STATES.ISSUED, customer=customer,
             transaction_currency='RON')


@pytest.fixture()
def invoice(customer, proforma):
    return G(Invoice, related_document=proforma, state=Invoice.STATES.ISSUED,
             customer=customer, transaction_currency='RON')


@pytest.fixture()
def transaction(db, customer, payment_processor, payment_method, proforma, invoice):
    return G(Transaction, invoice=invoice, proforma=proforma, currency='RON',
             amount=invoice.total, payment_method=payment_method)


@pytest.fixture()
def transaction_triggered(customer, payment_method_triggered, proforma, invoice):
    return G(Transaction, invoice=invoice, proforma=proforma, currency='RON',
             amount=invoice.total, payment_method=payment_method_triggered)


@pytest.fixture()
def transaction_triggered_v2(customer, payment_method_triggered_v2, proforma, invoice):
    return G(Transaction, invoice=invoice, proforma=proforma, currency='RON',
             amount=invoice.total, payment_method=payment_method_triggered_v2)



from datetime import datetime

import pytz
from django.utils.six import text_type

from payu.forms import PayULiveUpdateForm

from django import forms

from silver.payment_processors.forms import GenericTransactionForm
from silver.utils.payments import get_payment_complete_url
from silver.utils.international import countries


class PayUTransactionFormBase(GenericTransactionForm, PayULiveUpdateForm):
    def __init__(self, payment_method, transaction, billing_details,
                 request=None, *args, **kwargs):

        kwargs['initial'] = self._build_form_body(transaction, request)
        kwargs['initial'].update(billing_details)

        super(PayUTransactionFormBase, self).__init__(payment_method,
                                                      transaction,
                                                      request, **kwargs)

    def _build_form_body(self, transaction, request):
        return {
            'ORDER_REF': str(transaction.uuid),
            'ORDER_DATE':  datetime.now(pytz.UTC).strftime('%Y-%m-%d %H:%M:%S'),
            'PRICES_CURRENCY': transaction.currency,
            'CURRENCY': transaction.currency,
            'PAY_METHOD': 'CCVISAMC',
            'AUTOMODE': '1',
            'BACK_REF': get_payment_complete_url(transaction, request),
            'ORDER': self._get_order(transaction)
        }

    def _get_order(self, transaction):
        document = transaction.document
        product_name = text_type('Payment for {} {}-{}').format(document.kind,
                                                                document.series,
                                                                document.number)
        return [{
            'PNAME': product_name,
            'PCODE': text_type('{}-{}').format(document.series, document.number),
            'PRICE': text_type(transaction.amount),
            'PRICE_TYPE': 'GROSS',
            'VAT': text_type(document.sales_tax_percent or '0')
        }]


class PayUTransactionFormManual(PayUTransactionFormBase):
    pass


class PayUTransactionFormTriggered(PayUTransactionFormBase):
    def _build_form_body(self, transaction, request):
        form_body = super(PayUTransactionFormTriggered, self)._build_form_body(
            transaction, request
        )

        form_body['LU_ENABLE_TOKEN'] = '1'

        return form_body


class PayUTransactionFormTriggeredV2(PayUTransactionFormBase):
    def _build_form_body(self, transaction, request):
        form_body = super(PayUTransactionFormTriggeredV2, self)._build_form_body(
            transaction, request
        )

        form_body['LU_ENABLE_TOKEN'] = '1'
        form_body['LU_TOKEN_TYPE'] = 'PAY_ON_TIME'

        return form_body


class PayUBillingForm(GenericTransactionForm):
    email = forms.EmailField()
    first_name = forms.CharField()
    last_name = forms.CharField()
    phone = forms.CharField()
    city = forms.CharField()
    country = forms.ChoiceField(choices=countries)

    def __init__(self, payment_method, transaction,
                 request=None, data=None, *args, **kwargs):

        if not data:
            data = self._build_form_body(payment_method.customer)

        super(PayUBillingForm, self).__init__(payment_method, transaction,
                                              request, data, **kwargs)

    def to_payu_billing(self):
        data = self.cleaned_data
        return {
            'BILL_FNAME': data['first_name'],
            'BILL_LNAME': data['last_name'],
            'BILL_EMAIL': data['email'],
            'BILL_PHONE': data['phone'],
            'BILL_CITY': data['city'],
            'BILL_COUNTRYCODE': data['country']
        }

    def _build_form_body(self, customer):
        form_body = {
            'first_name': customer.first_name,
            'last_name': customer.last_name,
            'email': customer.email,
            'phone': customer.phone or '',
            'country': customer.country,
            'city': customer.city
        }

        return form_body



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from xml.etree import ElementTree

from django.db import transaction as django_transaction
from django.dispatch import receiver
from django.utils.six import text_type
from django_fsm import TransitionNotAllowed
from payu.payments import ALUPayment, TokenPayment
from payu.signals import payment_authorized, alu_token_created, payment_completed
from silver.models import Transaction
from silver.payment_processors import PaymentProcessorBase
from silver.payment_processors.mixins import (TriggeredProcessorMixin,
                                              ManualProcessorMixin)

from silver_payu.errors import TOKEN_ERROR_CODES, ALU_ERROR_CODES
from silver_payu.forms import (PayUTransactionFormManual,
                               PayUTransactionFormTriggered, PayUBillingForm,
                               PayUTransactionFormTriggeredV2)
from silver_payu.models import PayUPaymentMethod
from silver_payu.views import PayUTransactionView


class PayUBase(PaymentProcessorBase):
    payment_method_class = PayUPaymentMethod
    transaction_view_class = PayUTransactionView
    allowed_currencies = ('RON', 'USD', 'EUR')

    _has_been_setup = False

    def get_form(self, transaction, request):
        form = PayUBillingForm(payment_method=transaction.payment_method,
                               transaction=transaction, request=request,
                               data=request.POST)

        if form.is_valid():
            if not transaction.payment_method.archived_customer:
                address = text_type("{} {}").format(transaction.document.customer.address_1,
                                                    transaction.document.customer.address_2)

                archived_customer = form.to_payu_billing()
                archived_customer['BILL_ADDRESS'] = address

                transaction.payment_method.archived_customer = archived_customer
                transaction.payment_method.save()

            form = self.form_class(payment_method=transaction.payment_method,
                                   transaction=transaction, request=request,
                                   billing_details=form.to_payu_billing())

        return form

    def refund_transaction(self, transaction, payment_method=None):
        pass

    def void_transaction(self, transaction, payment_method=None):
        pass

    def handle_transaction_response(self, transaction, request):
        with django_transaction.atomic():
            Transaction.objects.select_for_update().filter(pk=transaction.pk).get()
            transaction.refresh_from_db()

            try:
                if request.GET.get('ctrl', None):
                    transaction.data['ctrl'] = request.GET['ctrl']
                    transaction.process()
                else:
                    error = request.GET.get('err', None) or 'Unknown error'
                    transaction.fail(fail_reason=error)
            except TransitionNotAllowed:
                pass

            transaction.save()


class PayUManual(PayUBase, ManualProcessorMixin):
    template_slug = 'payu_manual'
    form_class = PayUTransactionFormManual


class PayUTriggered(PayUBase, TriggeredProcessorMixin):
    """
    Uses TokenV1 API for recurrent payments.
    """
    template_slug = 'payu_triggered'
    form_class = PayUTransactionFormTriggered

    def execute_transaction(self, transaction):
        """
        :param transaction: A PayU transaction in Initial or Pending state.
        :return: True on success, False on failure.
        """

        if transaction.state != Transaction.States.Pending:
            return False

        return self._charge_transaction(transaction)

    def _charge_transaction(self, transaction):
        token = transaction.payment_method.token

        billing_details = transaction.payment_method.archived_customer
        try:
            delivery_details = {
                "DELIVERY_ADDRESS": billing_details["BILL_ADDRESS"],
                "DELIVERY_CITY": billing_details["BILL_CITY"],
                "DELIVERY_EMAIL": billing_details["BILL_EMAIL"],
                "DELIVERY_FNAME": billing_details["BILL_FNAME"],
                "DELIVERY_LNAME": billing_details["BILL_LNAME"],
                "DELIVERY_PHONE": billing_details["BILL_PHONE"]
            }
        except KeyError as error:
            transaction.fail(fail_reason=text_type('Invalid customer details. [{}]').format(error))
            transaction.save()
            return False

        payment_details = {
            "AMOUNT": str(transaction.amount),
            "CURRENCY": str(transaction.currency),
            "EXTERNAL_REF": str(transaction.uuid)
        }
        payment_details.update(billing_details)
        payment_details.update(delivery_details)

        payment = TokenPayment(payment_details, token)

        try:
            result = payment.pay()
        except Exception as error:
            transaction.fail(fail_reason=text_type(error))
            transaction.save()
            return False

        return self._parse_result(transaction, result)

    def _parse_result(self, transaction, result):
        try:
            result = json.loads(result)

            if "code" in result and not int(result["code"]):
                return True

            error_code, error_reason = self._parse_response_error(result)
            transaction.fail(fail_code=error_code, fail_reason=error_reason)
        except ValueError as error:
            transaction.fail(fail_reason=text_type(error))

        transaction.save()

        return False

    def _parse_response_error(self, payu_response):
        if not isinstance(payu_response, dict) or 'code' not in payu_response:
            return 'default', text_type('Missing payu error code.({})').format(payu_response)

        if str(payu_response['code']) in TOKEN_ERROR_CODES:
            error = TOKEN_ERROR_CODES[str(payu_response['code'])]
            return error['silver_code'], error['reason']

        return 'default', text_type('Unknown error code {}').format(payu_response['code'])


class PayUTriggeredV2(PayUBase, TriggeredProcessorMixin):
    """
    Uses ALUPaymentV3 API for recurrent payments.
    """
    template_slug = 'payu_triggered'
    form_class = PayUTransactionFormTriggeredV2

    def execute_transaction(self, transaction):
        """
        :param transaction: A PayU transaction in Initial or Pending state.
        :return: True on success, False on failure.
        """

        if transaction.state != Transaction.States.Pending:
            return False

        return self._charge_transaction(transaction)

    def _charge_transaction(self, transaction):
        token = transaction.payment_method.token

        customer_details = transaction.payment_method.archived_customer
        try:
            delivery_details = {
                "DELIVERY_ADDRESS": customer_details["BILL_ADDRESS"],
                "DELIVERY_CITY": customer_details["BILL_CITY"],
                "DELIVERY_EMAIL": customer_details["BILL_EMAIL"],
                "DELIVERY_FNAME": customer_details["BILL_FNAME"],
                "DELIVERY_LNAME": customer_details["BILL_LNAME"],
                "DELIVERY_PHONE": customer_details["BILL_PHONE"]
            }
        except KeyError as error:
            transaction.fail(fail_reason=text_type('Invalid customer details. [{}]').format(error))
            transaction.save()
            return False

        payment_details = {
            "PRICES_CURRENCY": str(transaction.currency),
            "ORDER_REF": str(transaction.uuid),
            "PAY_METHOD": "CCVISAMC",
        }

        if transaction.document:
            pname = text_type("{provider} {doc_type} {doc_number}").format(
                doc_type=transaction.document.kind,
                doc_number=transaction.document.series_number,
                provider=transaction.provider.name
            )

            vat = str(int(transaction.document.sales_tax_percent))
            price_type = "GROSS"  # (VAT included)
        else:
            pname = text_type("Payment for {provider}").format(provider=transaction.provider.name)

            vat = "0"
            price_type = "NET"  # (VAT will be added by PayU)

        order_details = [
            {
                'PNAME': pname,
                'PCODE': str(transaction.uuid),
                'PRICE': str(transaction.amount),
                'VAT': vat,
                'PRICE_TYPE': price_type,
                'QTY': '1',
            }
        ]

        payment_details.update(customer_details)
        payment_details.update(delivery_details)
        payment_details["ORDER"] = order_details

        payment = ALUPayment(payment_details, token, lu_token_type="PAY_ON_TIME")

        try:
            result = payment.pay()
        except Exception as error:
            transaction.fail(fail_reason=text_type(error))
            transaction.save()
            return False

        return self._parse_result(transaction, result)

    def _parse_result(self, transaction, result):
        try:
            element = ElementTree.fromstring(result)
            status = element.find("STATUS").text
            return_code = element.find('RETURN_CODE').text

            if status == "SUCCESS":
                return True

            error_code, error_reason = self._parse_response_error(return_code)
            transaction.data = {
                "status": status,
                "message": error_reason,
                "return_code": return_code,
            }

            return_message = element.find("RETURN_MESSAGE")
            if return_message is not None:
                transaction.data["return_message"] = return_message.text

            transaction.fail(fail_code=error_code, fail_reason=error_reason)
        except ValueError as error:
            transaction.fail(fail_reason=text_type(error))

        transaction.save()

        return False

    def _parse_response_error(self, return_code):
        if str(return_code) in ALU_ERROR_CODES:
            error = ALU_ERROR_CODES[str(return_code)]
            return error['silver_code'], error['reason']

        return 'default', text_type('Unknown error code {}').format(return_code)


@receiver([payment_authorized, payment_completed])
def payu_ipn_received(sender, **kwargs):
    transaction = Transaction.objects.get(uuid=sender.REFNOEXT)

    try:
        if transaction.state != Transaction.States.Settled:
            transaction.settle()
            transaction.save()
    except TransitionNotAllowed as error:
        try:
            transaction.fail(fail_reason=text_type(error))
            transaction.save()
        except TransitionNotAllowed:
            transaction.fail_reason = text_type(error)
            transaction.save()

        raise


@receiver(alu_token_created)
def payu_token_received(sender, **kwargs):
    transaction = Transaction.objects.get(uuid=sender.ipn.REFNOEXT)
    payment_method = PayUPaymentMethod.objects.get(pk=transaction.payment_method_id)

    payment_processor = payment_method.get_payment_processor()
    if payment_processor.__class__ is PayUTriggered:
        payment_method.token = sender.IPN_CC_TOKEN
    elif payment_processor.__class__ is PayUTriggeredV2:
        if not sender.TOKEN_HASH:
            return
        payment_method.token = sender.TOKEN_HASH
    else:
        # no other PayU payment processor implementation expects tokens
        return

    payment_method.verified = True
    payment_method.display_info = sender.IPN_CC_MASK
    payment_method.valid_until = sender.IPN_CC_EXP_DATE
    payment_method.save()



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

from django_fsm import transition

from silver.models import PaymentMethod


class PayUPaymentMethod(PaymentMethod):
    class Meta:
        proxy = True

    @property
    def token(self):
        return self.decrypt_data(self.data.get('token'))

    @token.setter
    def token(self, value):
        self.data['token'] = self.encrypt_data(value)

    @property
    def archived_customer(self):
        raw_customer = self.data.get('archived_customer', '')
        return json.loads(self.decrypt_data(raw_customer) or '{}')

    @archived_customer.setter
    def archived_customer(self, value):
        raw_customer = json.dumps(value)
        self.data['archived_customer'] = self.encrypt_data(raw_customer)



import pytest
from faker import Faker
from mock import patch, MagicMock

from silver_payu.forms import PayUTransactionFormBase, PayUBillingForm

from .fixtures import *


faker = Faker()


@patch('silver_payu.forms.datetime')
def test_payu_transaction_form_build_body(mocked_datetime, transaction, payment_method):
    mocked_datetime.now.return_value.strftime.return_value = 'order_date'

    with patch('silver.utils.payments._get_jwt_token') as mocked_token:
        mocked_token.return_value = 'token'

        form = PayUTransactionFormBase(payment_method, transaction, {})

        assert form._build_form_body(transaction, None) == {
            'AUTOMODE': '1',
            'BACK_REF': '/pay/token/complete',
            'CURRENCY': 'RON',
            'ORDER': [{'PCODE': '%s-%s' % (transaction.document.series,
                                           transaction.document.number),
                       'PNAME': 'Payment for %s %s-%s' % (transaction.document.kind,
                                                          transaction.document.series,
                                                          transaction.document.number),
                       'PRICE': str(transaction.amount),
                       'PRICE_TYPE': 'GROSS',
                       'VAT': str(transaction.document.sales_tax_percent)}],
            'ORDER_DATE': 'order_date',
            'ORDER_REF': '%s' % transaction.uuid,
            'PAY_METHOD': 'CCVISAMC',
            'PRICES_CURRENCY': 'RON'
        }


@pytest.mark.parametrize('data, form_class, archived_customer', [
    ({
        'email': faker.email(),
        'first_name': faker.first_name(),
        'last_name': faker.last_name(),
        'phone': faker.phone_number(),
        'country': '',
        'fiscal_code': ''
    }, PayUBillingForm, {}),
    ({
        'email': 'john@acme.com',
        'first_name': 'John',
        'last_name': 'Doe',
        'phone': '+40000000000',
        'country': 'RO',
        'city': 'Timisoara',
        'fiscal_code': ''
    }, PayUTransactionFormBase, {
        'BILL_ADDRESS': u"Zalaegerszegi utca 65.\nH-5484 ligetv\xc3\xa1r 9",
        'BILL_CITY': 'Timisoara',
        'BILL_COUNTRYCODE': 'RO',
        'BILL_EMAIL': 'john@acme.com',
        'BILL_FNAME': 'John',
        'BILL_LNAME': 'Doe',
        'BILL_PHONE': '+40000000000'
    })
])
def test_payment_processor_get_form(payment_processor, transaction, data,
                                    form_class, archived_customer):
    form = payment_processor.get_form(transaction, MagicMock(POST=data))

    assert isinstance(form, form_class)
    assert transaction.payment_method.archived_customer == archived_customer



TOKEN_ERROR_CODES = {
    '300': {
        'silver_code': 'default',
        'reason': 'The REF_NO specified is not a valid transaction',
    },

    '400': {
        'silver_code': 'default',
        'reason': 'The METHOD variable needs to have one of the following values:TOKEN_NEWSALE, TOKEN_CANCEL, TOKEN_GETINFO',
    },

    '500': {
        'silver_code': 'default',
        'reason': 'The value of TIMESTAMP differs too much from the current time. Check your system clock and ensure that TIMESTAMP is in UTC timezone.',
    },

    '600': {
        'silver_code': 'default',
        'reason': 'Make sure that your MerchantID is the same one found in yourPayU Control Panel',
    },

    '601': {
        'silver_code': 'insufficient_funds',
        'reason': 'The credit card used has insufficient funds available.',
    },

    '602': {
        'silver_code': 'expired_card',
        'reason': 'The credit card used is expired',
    },

    '603': {
        'silver_code': 'default',
        'reason': 'Temporary processing error. Retrying after a few minutes should work.',
    },

    '604': {
        'silver_code': 'invalid_card',
        'reason': 'The credit card used is invalid.',
    },

    '605': {
        'silver_code': 'default',
        'reason': 'General system error. Retrying after a few minutes should work but if it\'s not please contact our support team.',
    },

    '606': {
        'silver_code': 'transaction_declined_by_bank',
        'reason': 'Invalid Transaction error specified by the credit card company.',
    },

    '607': {
        'silver_code': 'default',
        'reason': 'The bank is still processing the transaction, check order status using IOS webservice.',
    },

    '1200': {
        'silver_code': 'default',
        'reason': 'There is a problem with your SIGN variable. Please check your code.',
    },

    '1300': {
        'silver_code': 'default',
        'reason': 'The REF_NO you have specified is not valid. Please check the value.',
    },

    '1500': {
        'silver_code': 'default',
        'reason': 'Invalid Token command for the REF_NO you have specified.',
    },

    '1600': {
        'silver_code': 'default',
        'reason': 'Invalid External Ref No',
    },

    '1900': {
        'silver_code': 'default',
        'reason': 'The AMOUNT value should be a positive number, either integer or a float.',
    },

    '2000': {
        'silver_code': 'default',
        'reason': 'You have exceeded the maximum amount limit for your terminal. Please try again in a few minutes.',
    },

    '2100': {
        'silver_code': 'default',
        'reason': 'CURRENCY variable has an unsupported or invalid value.',
    },

    '2200': {
        'silver_code': 'expired_payment_method',
        'reason': 'Operation was not performed because the token has expired.',
    },

    '2300': {
        'silver_code': 'expired_payment_method',
        'reason': 'Operation was not performed because the token has expired.',
    },

    '2401': {
        'silver_code': 'default',
        'reason': 'BILL_LNAME field is mandatory',
    },

    '2402': {
        'silver_code': 'default',
        'reason': 'BILL_FNAME field is mandatory',
    },

    '2403': {
        'silver_code': 'default',
        'reason': 'BILL_EMAIL field is mandatory',
    },

    '2404': {
        'silver_code': 'default',
        'reason': 'BILL_EMAIL field is not a valid e-mail',
    },

    '2405': {
        'silver_code': 'default',
        'reason': 'BILL_PHONE field is mandatory',
    },

    '2406': {
        'silver_code': 'default',
        'reason': 'BILL_ADDRESS field is mandatory',
    },

    '2407': {
        'silver_code': 'default',
        'reason': 'BILL_CITY field is mandatory',
    },

    '2408': {
        'silver_code': 'default',
        'reason': 'DELIVERY_LNAME field is mandatory',
    },

    '2409': {
        'silver_code': 'default',
        'reason': 'DELIVERY_FNAME field is mandatory',
    },

    '2410': {
        'silver_code': 'default',
        'reason': 'DELIVERY_PHONE field is mandatory',
    },

    '2411': {
        'silver_code': 'default',
        'reason': 'DELIVERY_ADDRESS field is mandatory',
    },

    '2412': {
        'silver_code': 'default',
        'reason': 'DELIVERY_CITY field is mandatory',
    },

    '2413': {
        'silver_code': 'default',
        'reason': 'DELIVERY_EMAIL field is not a valid e-mail',
    },

    '2414': {
        'silver_code': 'default',
        'reason': 'BILL_COUNTRYCODE field is not a valid ISO 3166-1 alpha-2country code',
    },

    '2415': {
        'silver_code': 'default',
        'reason': 'DELIVERY_COUNTRYCODE field is not a valid ISO 3166-1 alpha-2country code',
    },

    '3000': {
        'silver_code': 'default',
        'reason': 'Invalid number of installments selected',
    },

    '3100': {
        'silver_code': 'default',
        'reason': 'Token is already persistent',
    },

    '3101': {
        'silver_code': 'expired_payment_method',
        'reason': 'Token has expired',
    },

    '3200': {
        'silver_code': 'default',
        'reason': 'Token usage on marketplace orders is restricted on this protocol',
    },

    '4000': {
        'silver_code': 'default',
        'reason': 'Please check if the card scheme used to make the original transaction is still enabled for that merchant.',
    },
}

ALU_ERROR_CODES = {
    "ALREADY_AUTHORIZED": {
        "silver_code": "default",
        "reason": "Tried to place a new order with the same ORDER_REF and HASH as previous one."
    },
    "AUTHORIZATION_FAILED": {
        "silver_code": "default",
        "reason": "The payment was not authorized."
    },
    "INVALID_CUSTOMER_INFO": {
        "silver_code": "default",
        "reason": "Required data from the Shopper is missing or is malformed."
    },
    "INVALID_PAYMENT_INFO": {
        "silver_code": "default",
        "reason": "Card data is not correct."
    },
    "INVALID_ACCOUNT": {
        "silver_code": "default",
        "reason": "The Merchant name is not correct."
    },
    "INVALID_PAYMENT_METHOD_CODE": {
        "silver_code": "default",
        "reason": "Payment method code is NOT recognized."
    },
    "INVALID_CURRENCY": {
        "silver_code": "default",
        "reason": "Payment currency is not recognized."
    },
    "REQUEST_EXPIRED": {
        "silver_code": "default",
        "reason": "The request has expired based on provided ORDER_DATE."
    },
    "HASH_MISMATCH": {
        "silver_code": "default",
        "reason": "Hash sent by the Merchant does not match the hash calculated by PayU."
    },
    "WRONG_VERSION": {
        "silver_code": "default",
        "reason": "ALU version sent by the Merchant does not exist."
    },
    "INVALID_CC_TOKEN": {
        "silver_code": "default",
        "reason": "CC_TOKEN sent by the Merchant is not valid."
    },
}



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from django.conf.urls import include, url

from silver.views import pay_transaction_view, complete_payment_view


urlpatterns = [
    url(r'^', include('payu.urls')),

    url(r'pay/(?P<token>[0-9a-zA-Z-_\.]+)/$',
        pay_transaction_view, name='payment'),
    url(r'pay/(?P<token>[0-9a-zA-Z-_\.]+)/complete$',
        complete_payment_view, name='payment-complete')
]



import pytest
from faker import Faker
from mock import MagicMock, patch
from django_dynamic_fixture import G

from django.utils.dateparse import parse_datetime

from silver.models import Transaction, PaymentMethod
from silver_payu.forms import PayUBillingForm, PayUTransactionFormBase
from silver_payu.models import PayUPaymentMethod
from silver_payu.payment_processors import payu_ipn_received, payu_token_received

from .fixtures import *

faker = Faker()


def test_payment_method_data_set():
    payment_method = PayUPaymentMethod()

    payment_method.decrypt_data = lambda value: value
    payment_method.encrypt_data = lambda value: value

    payment_method.token = "random token"
    assert payment_method.token == "random token"

    payment_method.archived_customer = {'name': 'test'}
    assert payment_method.archived_customer == {'name': 'test'}


@pytest.mark.django_db
def test_execute_transaction_wrong_payment_processor(payment_processor_triggered,
                                                     transaction_triggered):
    assert not payment_processor_triggered.execute_transaction(transaction_triggered)


@pytest.mark.django_db
def test_execute_transaction_wrong_transaction_state(payment_processor_triggered):
    payment_processor_triggered._charge_transaction = lambda x: True

    transaction_triggered = MagicMock(payment_processor=payment_processor_triggered,
                                      state=Transaction.States.Settled)
    assert not payment_processor_triggered.execute_transaction(transaction_triggered)


@pytest.mark.django_db
def test_execute_transaction_happy_path(payment_processor_triggered):
    payment_processor_triggered._charge_transaction = lambda x: True

    transaction_triggered = MagicMock(payment_processor=payment_processor_triggered,
                                      state=Transaction.States.Pending)
    assert payment_processor_triggered.execute_transaction(transaction_triggered)


@pytest.mark.django_db
@patch('silver_payu.payment_processors.TokenPayment')
def test_charge_transaction_triggered(mocked_token_payment,
                                      payment_processor_triggered,
                                      payment_method, transaction_triggered):
    mocked_token_payment.return_value.pay.return_value = '{"code": "0"}'

    payment_method.token = faker.word()
    payment_method.archived_customer = {
        "BILL_ADDRESS": faker.address(),
        "BILL_CITY": faker.city(),
        "BILL_EMAIL": faker.email(),
        "BILL_FNAME": faker.first_name(),
        "BILL_LNAME": faker.last_name(),
        "BILL_PHONE": faker.phone_number(),
    }

    assert payment_processor_triggered._charge_transaction(transaction_triggered)

    asserted_payment_details = payment_method.archived_customer
    asserted_payment_details.update({
        "DELIVERY_ADDRESS": asserted_payment_details["BILL_ADDRESS"],
        "DELIVERY_CITY": asserted_payment_details["BILL_CITY"],
        "DELIVERY_EMAIL": asserted_payment_details["BILL_EMAIL"],
        "DELIVERY_FNAME": asserted_payment_details["BILL_FNAME"],
        "DELIVERY_LNAME": asserted_payment_details["BILL_LNAME"],
        "DELIVERY_PHONE": asserted_payment_details["BILL_PHONE"],
        "AMOUNT": str(transaction_triggered.amount),
        "CURRENCY": str(transaction_triggered.currency),
        "EXTERNAL_REF": str(transaction_triggered.uuid),
    })

    mocked_token_payment.assert_called_once_with(asserted_payment_details,
                                                 payment_method.token)


@pytest.mark.django_db
@pytest.mark.parametrize('payment_result, excepted_return', [
    ('{"code": "0"}', True),
    ('{"code": "1"}', False),
    ('{code: "1"}', False),
])
def test_parse_token_payment_result(payment_processor_triggered,
                                    transaction_triggered, payment_result,
                                    excepted_return):
    assert payment_processor_triggered._parse_result(transaction_triggered,
                                                     payment_result) == excepted_return


@pytest.mark.django_db
@patch('silver.models.transactions.transaction.Transaction.update_document_state')
def test_ipn_received(mocked_document, transaction):
    payu_ipn_received(MagicMock(REFNOEXT=transaction.uuid))

    transaction.refresh_from_db()

    assert transaction.state == "settled"


@pytest.mark.django_db
def test_token_received(transaction_triggered):
    sender = MagicMock(ipn=MagicMock(REFNOEXT=transaction_triggered.uuid),
                       IPN_CC_TOKEN=faker.word(), IPN_CC_EXP_DATE='2017-07-31',
                       IPN_CC_MASK=faker.word())
    payu_token_received(sender)

    transaction_triggered.payment_method.refresh_from_db()

    expected_valid_until = parse_datetime(sender.IPN_CC_EXP_DATE + " 00:00:00")
    assert transaction_triggered.payment_method.valid_until == expected_valid_until
    assert transaction_triggered.payment_method.token == sender.IPN_CC_TOKEN
    assert transaction_triggered.payment_method.display_info == sender.IPN_CC_MASK
    assert transaction_triggered.payment_method.verified


@pytest.mark.django_db
def test_token_v2_received(transaction_triggered_v2):
    sender = MagicMock(ipn=MagicMock(REFNOEXT=transaction_triggered_v2.uuid),
                       IPN_CC_TOKEN=faker.word(), TOKEN_HASH=faker.word(),
                       IPN_CC_EXP_DATE='2017-07-31', IPN_CC_MASK=faker.word())
    payu_token_received(sender)

    transaction_triggered_v2.payment_method.refresh_from_db()

    expected_valid_until = parse_datetime(sender.IPN_CC_EXP_DATE + " 00:00:00")
    assert transaction_triggered_v2.payment_method.valid_until == expected_valid_until
    assert transaction_triggered_v2.payment_method.token == sender.TOKEN_HASH
    assert transaction_triggered_v2.payment_method.display_info == sender.IPN_CC_MASK
    assert transaction_triggered_v2.payment_method.verified
